!=========================================================================
! Copyright (C) GemTalk Systems 1986-2016.  All Rights Reserved.
!
! $Id: rcindexdictionary.gs,v 1.10 2008-01-09 19:02:36 stever Exp $
!
! Superclass Hierarchy:
!   RcIndexDictionary, RcKeyValueDictionary, AbstractDictionary,
!   Collection, Object.
!
!=========================================================================

! class created in idxclasses.topaz

removeallmethods RcIndexDictionary
removeallclassmethods RcIndexDictionary
protectmethods

category: 'For Documentation Installation only'
classmethod: RcIndexDictionary
installDocumentation

< unprotected >

| doc |
doc := GsClassDocumentation _newForPrivateGsClass: self.

doc documentClassWith: (GsDocText new).

self description: doc.
%

! ------------------- Class methods for RcIndexDictionary
category: 'Instance Creation'
classmethod: RcIndexDictionary
new

"Returns a RcIndexDictionary with the default table size."
<protected>

^ self new: self defaultBasicSize
%

category: 'Constants'
classmethod: RcIndexDictionary
defaultBasicSize

"Returns the default basic size."
<unprotected>

^ 2003
%

! ------------------- Instance methods for RcIndexDictionary
category: 'Removing'
method: RcIndexDictionary
_removeKey: aKey value: aValue term: pathTerm

"Removes the key/value/pathTerm triplet from the receiver.  Returns true if the
 removal was successful, false if not."

< unprotected primitive: 901>
| collisionBkt |
collisionBkt := self indexBucketFor: aKey.

collisionBkt
    removeKey: aKey
    value: aValue
    term: pathTerm
    root: self
    logging: false
    ifAbsent: [ System _disableProtectedMode. ^ false ].

System _disableProtectedMode.
^ true
%

category: 'Updating'
method: RcIndexDictionary
at: aKey put: aValue term: pathTerm

"Stores the aKey/aValue/pathTerm triplet in the index dictionary.  Returns the
 value."
<protected>

self _at: aKey put: aValue term: pathTerm logging: true.
^ aValue
%

category: 'Accessing'
method: RcIndexDictionary
at: aKey term: pathTerm ifAbsent: aBlock

"Returns the value that corresponds to aKey/pathTerm.  If no such
 key/value/pathTerm triplet exists, returns the result of evaluating the
 zero-argument block aBlock."
<protected>

^ (self indexBucketFor: aKey) at: aKey term: pathTerm ifAbsent: aBlock
%

category: 'Constants'
method: RcIndexDictionary
collisionLimitPerBucket

"Returns the number of entries in a collision bucket such that if a greater
 number is inserted, the hash dictionary will grow.  This number is chosen such
 that a collision bucket will occupy a single page
 (collisionLimitPerBucket * entrySize <= 2000)."
<unprotected>

^ 660
%

category: 'Initializing'
method: RcIndexDictionary
initializeWorkingSetInterval: indexObj

"Determines the working set interval (the number of collision buckets that are
 considered for writing during index creation).  If the instance variable
 'okToUpdate' is set to true, then index creation does not control the working
 set (and the working set interval is not used)."

<protected>

" indicate whether it is not ok to update the dictionary
  (so look at working set instead) "
" only use this algorithm if NSC size is greater than the working
  set threshold (default is 1MByte) "
okToUpdate := indexObj nscRoot size < (self getWorkingSetThreshold: indexObj nscRoot).

workingSetInterval := self getWorkingSetInterval: indexObj nscRoot.

okToUpdate
  ifFalse: [ | array |
    array := Array new: (self _basicSize // workingSetInterval).
    1 to: array size do: [ :i |
      array at: i put: (IndexDictionaryEntryHolder new indexDictionary: self)
    ].
    entryHolders := array.
  ]
%

category: 'Initializing'
method: RcIndexDictionary
initialize: itsSize

"Initializes the receiver and sets its collisionLimitPerBucket."

<protected>

super initialize: itsSize.
collisionLimitPerBucket := self collisionLimitPerBucket.
%

category: 'Private'
method: RcIndexDictionary
logAt: aKey put: aValue term: pathTerm inCollisionBucket: aCollisionBkt

"Create a log entry for adding the given key, value, and pathTerm.  The
 collision bucket is the object on which a conflict may occur."
<protected>

| logEntry |
logEntry := LogEntry new.
logEntry receiver: self;
    selector: #_at:put:term:logging:;
    argArray: #[ aKey, aValue, pathTerm, false ].
System redoLog addLogEntry: logEntry forConflictObject: aCollisionBkt
%

category: 'Private'
method: RcIndexDictionary
logRemoveKey: aKey value: aValue term: pathTerm
  inCollisionBucket: aCollisionBkt

"Creates a log entry for removing the given key/value/pathTerm triplet.  The
 collision bucket is the object on which a conflict may occur."
<protected>

| logEntry |
logEntry := LogEntry new.
logEntry receiver: self;
    selector: #_removeKey:value:term:;
    argArray: #[ aKey, aValue, pathTerm ].
System redoLog addLogEntry: logEntry forConflictObject: aCollisionBkt
%

category: 'Hashing'
method: RcIndexDictionary
rebuildTable

"Increases the size of the root object and rehashes the contents of the
 collision buckets."
<protected>

| newSize array collBkt constraintClass origSize begin |
" Let's not be so aggressive on table growth "
origSize := self _basicSize.
newSize := origSize * 2 - 1.

constraintClass := self class varyingConstraint.
array := Array new: origSize.
" store all buckets in an Array "
1 to: origSize do: [ :i |
  array at: i put: (self _at: i).
  self _at: i put: constraintClass new.
].

super _basicSize: newSize.
origSize + 1 to: newSize do: [ :i |
  self _at: i put: constraintClass new.
].

begin := (array _at: 1) _numCacheSlots + 1.

1 to: origSize do: [ :i |
  collBkt := array _at: i.
  begin to: begin + (collBkt numElements * 3) - 1 by: 3 do: 
    [ :j | 
    | jval |
    jval := (collBkt _at: j+2).
    (BucketValueBag _hasInstance: jval) 
      ifTrue: [
        jval do: 
          [:realValue |
          self _at: (collBkt _at: j)
            put: realValue
            term: (collBkt _at: j+1)
            logging: false
          ]
        ]
      ifFalse: [
        self _at: (collBkt _at: j)
          put: (collBkt _at: j+2)
          term: (collBkt _at: j+1)
          logging: false
        ]
     ].

  collBkt reset.
]
%

category: 'Removing'
method: RcIndexDictionary
removeKey: anObject ifAbsent: aBlock

"Disallowed.  Use removeKey:value:ifAbsent: instead."
<protected>

self shouldNotImplement: #removeKey:ifAbsent:
%

category: 'Removing'
method: RcIndexDictionary
removeKey: aKey
value: aValue
term: pathTerm
logging: aBoolean
ifAbsent: aBlock

"Remove the key/value/pathTerm triplet from the receiver.  If no
 key/value/pathTerm triplet is present, evaluate the zero-argument block
 aBlock.  Returns true if there are any other entries with the same key/term
 in the dictionary, and returns false otherwise."
<protected>

| collisionBkt more doLogging |
collisionBkt := self indexBucketFor: aKey.

doLogging := aBoolean.
more := collisionBkt
    removeKey: aKey
    value: aValue
    term: pathTerm
    root: self
    logging: doLogging
    ifAbsent: [
        " if entry not found, do not log the removal "
        doLogging := false.
        aBlock value.
    ].

doLogging
    ifTrue: [ " log the removal in the redo log "
        self logRemoveKey: aKey
            value: aValue
            term: pathTerm
            inCollisionBucket: collisionBkt
    ].

^ more
%

category: 'Removing'
method: RcIndexDictionary
removeKey: aKey value: aValue term: pathTerm logging: aBoolean

"Remove the key/value/pathTerm triplet from the receiver.  Returns true
 if there are any other entries with the same key/term in the dictionary,
 and returns false otherwise."
<protected>

| collisionBkt more doLogging |
collisionBkt := self indexBucketFor: aKey.
doLogging := aBoolean.

more := collisionBkt
    removeKey: aKey
    value: aValue
    term: pathTerm
    root: self
    logging: doLogging
    ifAbsent: [
        " if entry not found, do not log the removal "
        doLogging := false.
    ].

doLogging
    ifTrue: [ " log the removal in the redo log "
        self logRemoveKey: aKey
            value: aValue
            term: pathTerm
            inCollisionBucket: collisionBkt
    ].

^ more
%

category: 'Error Handling'
method: RcIndexDictionary
_errorNoEntryForKey: aKey term: aTerm value: aValue

"An entry for the key/term/value was not present in the dictionary."
<protected>

^ self
    _error: #rtErrIndexDictionaryEntryNotInDictionary
    args: #[ aKey, aTerm, aValue ]
%

category: 'Accessing'
method: RcIndexDictionary
changeToSegment: aSegment

"Assign the receiver and its index buckets the given segment."
<protected>

self assignToSegment: aSegment.
self _doCollisionBuckets: [ :indexBucket |
    indexBucket assignToSegment: aSegment
]
%

category: 'Accessing'
method: RcIndexDictionary
at: aKey term: pathTerm otherwise: aValue

"Returns the value that corresponds to aKey/pathTerm.  If no such
 key/value/pathTerm triplet exists, returns aValue."
<protected>

^ (self indexBucketFor: aKey) at: aKey term: pathTerm otherwise: aValue
%

category: 'Accessing'
method: RcIndexDictionary
firstAt: aKey term: pathTerm otherwise: aValue

"Returns the value that corresponds to aKey/pathTerm.  If no such
 key/value/pathTerm triplet exists, returns aValue.  If the value
 is a BucketValueBag, returns the first object in the Bag."
<protected>

^ (self indexBucketFor: aKey) firstAt: aKey term: pathTerm otherwise: aValue
%

category: 'Accessing'
method: RcIndexDictionary
at: aKey term: pathTerm

"Returns the value that corresponds to aKey/pathTerm.  If no such
 key/value/pathTerm triplet exists, raise an error."
<protected>

^ (self indexBucketFor: aKey) at: aKey term: pathTerm
%

category: 'Updating'
method: RcIndexDictionary
_existsAt: aKey put: aValue term: pathTerm logging: aBoolean

"Stores the aKey/aValue/pathTerm triplet in the index dictionary.  Rebuilds the
 hash table if the addition caused the number of collisions to exceed the limit
 allowed.  The argument aBoolean determines if an entry is placed in the redo
 log for this operation.  Returns whether the key/pathTerm pair already exists
 in the dictionary."
<protected>

| collisionBkt exists |
collisionBkt := self indexBucketFor: aKey.
exists := collisionBkt
  at: aKey
  put: aValue
  term: pathTerm
  root: self
  logging: aBoolean.

aBoolean
    ifTrue: [
        self logAt: aKey put: aValue term: pathTerm inCollisionBucket: collisionBkt
    ].

(collisionBkt size > collisionLimitPerBucket)
    ifTrue: [ self rebuildTable ].

^ exists
%

category: 'Query Support'
method: RcIndexDictionary
_findAllValuesNotIdenticalTo: aValue atTerm: pathTerm into: aCollection

"Iterate over all index buckets, finding all entries with the given path term.
 If the value at the path term is not identical to the given value, place it in
 the set."
<protected>

| collBkt val |
1 to: self _basicSize do: [ :i |
    collBkt := self _at: i.
    1 to: collBkt numElements do: [ :j |
        ( ( (collBkt termAt: j) == pathTerm ) _and:
        [ (collBkt keyAt: j) ~~ aValue ] )
            ifTrue: [
                val := collBkt valueAt: j.
                (collBkt _isBucketValueBagAt: (collBkt valueOffsetFor: j))
                    ifTrue: [ aCollection addAll: val ]
                    ifFalse: [ aCollection add: val ]
            ]
    ]
].
^ aCollection
%

category: 'Query Support'
method: RcIndexDictionary
_findFirstKeyNotIdenticalTo: aKey atTerm: pathTerm

"Iterates over the index buckets, finding the first entry with the given path
 term.  If the key at the path term is not identical to the given key, returns
 it.  If all key are identical, returns nil."
<protected>

| collBkt key |
1 to: self _basicSize do: [ :i |
    collBkt := self _at: i.
    1 to: collBkt numElements do: [ :j |
        ( ( (collBkt termAt: j) == pathTerm ) _and:
        [ (key := collBkt keyAt: j) ~~ aKey ] )
            ifTrue: [ ^ key ]
    ]
].
^ #_incompletePathTraversal
%

category: 'Accessing'
method: RcIndexDictionary
indexBucketFor: aKey

"Returns the RC index bucket for the given key."

<protected primitive: 164>
self _primitiveFailed: #indexBucketFor: .
self _uncontinuableError
%

category: 'Hashing'
method: RcIndexDictionary
hashFunction: aKey

"The hash function for aKey is calculated internally inside the method
 'indexBucketFor:'.  Raise a bogus error if this method is invoked."
<protected>

self _error: #rtErrNilKey
%

category: 'Accessing'
method: RcIndexDictionary
instVarAt: anIndex

"Protected version of Object | instVarAt:"

<protected primitive: 811"with rcvrClass">

| numInstVars |
(anIndex _isInteger)
ifTrue:
[
   numInstVars:= self class allInstVarNames size.
   (numInstVars == 0)
   ifTrue:
      [ ^ self _error: #rtErrNoInstVars].
   ((anIndex >= 1) & (anIndex <= numInstVars))
   ifFalse: "out of bounds"
      [ ^ self _errorIndexOutOfRange: anIndex]
]
ifFalse: "anIndex not an integer"
   [ ^ self _errorNonIntegerIndex: anIndex
].

self _primitiveFailed: #instVarAt: .
self _uncontinuableError
%

category: 'Accessing'
method: RcIndexDictionary
_at: anIndex

"Protected version of Object | _at:"

<protected primitive: 32>

(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
(anIndex _isInteger)
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex]
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].

self _primitiveFailed: #_at: .
self _uncontinuableError
%

category: 'Accessing'
method: RcIndexDictionary
_basicAt: anIndex

"Protected version of Object | _basicAt:"

<protected primitive: 32>

(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
(anIndex _isInteger)
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex]
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].

self _primitiveFailed: #_basicAt: .
self _uncontinuableError
%

category: 'Accessing'
method: RcIndexDictionary
_primitiveAt: anIndex

"Protected version of Object | _primitiveAt:"

<protected primitive: 89>

(anIndex _isInteger)
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex]
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].

self _primitiveFailed: #_primitiveAt: .
self _uncontinuableError
%

category: 'Updating'
method: RcIndexDictionary
instVarAt: anIndex put: aValue

"Protected version of Object | instVarAt:put:"

<protected primitive: 804"with rcvrClass">
(self class allInstVarNames size == 0)
  ifTrue: [ ^ self _error: #rtErrNoInstVars].
(anIndex _isInteger)
  ifTrue: [((anIndex < 1) | (anIndex > self class allInstVarNames size))
        ifTrue: [ ^ self _errorIndexOutOfRange: anIndex]] "out of bounds"
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].

self _primitiveFailed: #instVarAt:put: .
self _uncontinuableError
%

category: 'Updating'
method: RcIndexDictionary
_basicAt: anIndex put: aValue

"Protected version of Object | _basicAt:put:"

<protected primitive: 268>

(anIndex _isInteger)
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].
(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
((anIndex < 1) | (anIndex > (self size + 1))) "out of bounds"
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex].
(self class isBytes)
ifTrue:
  [ ((aValue _isSmallInteger not) _or: [ (aValue < 0) | (aValue > 255) ])
    ifTrue: [^ aValue _error: #rtErrExpectedByteValue].
  ].

self _primitiveFailed: #_basicAt:put: .
self _uncontinuableError
%

category: 'Updating'
method: RcIndexDictionary
_at: anIndex put: aValue

"Protected version of Object | _at:put:"

<protected primitive: 268>

(anIndex _isInteger)
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].
(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
((anIndex < 1) | (anIndex > (self size + 1))) "out of bounds"
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex].
(self class isBytes)
ifTrue:
  [ ((aValue _isSmallInteger not) _or: [ (aValue < 0) | (aValue > 255) ])
    ifTrue: [^ aValue _error: #rtErrExpectedByteValue].
  ].

self _primitiveFailed: #_at:put: .
self _uncontinuableError
%

category: 'Updating'
method: RcIndexDictionary
_basicSize: anInteger

"Protected version of Object | _basicSize:"

<protected primitive: 803"with rcvrClass">

(self class isIndexable not) "not an indexable object"
  ifTrue: [ ^ self _errorNotIndexable].
(anInteger _isInteger)
   ifTrue: [ ^ self _errorIndexOutOfRange: anInteger]
   ifFalse: [ ^ self _errorNonIntegerIndex: anInteger].
self _primitiveFailed: #_basicSize: .
self _uncontinuableError
%

category: 'Updating'
method: RcIndexDictionary
_primitiveAt: anIndex put: aValue

"Protected version of Object | _primitiveAt:put:"

<protected primitive: 281>

(anIndex _isInteger)
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].
((anIndex < 1) | (anIndex > (self _primitiveSize + 1))) "out of bounds"
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex].
(self class isBytes)
ifTrue:
  [ ((aValue _isSmallInteger not) _or: [ (aValue < 0) | (aValue > 255) ])
    ifTrue: [^ aValue _error: #rtErrExpectedByteValue].
  ].

self _primitiveFailed: #_primitiveAt:put: .
self _uncontinuableError
%

category: 'Reduced Conflict Support'
method: RcIndexDictionary
_rcAt: anIndex

"Protected version of Object | _rcAt:"

<protected primitive: 92>
self _primitiveFailed: #_rcAt: .
self _uncontinuableError
%

category: 'Accessing'
method: RcIndexDictionary
basicAt: anIndex

"Protected version of Object | basicAt:"
<protected>

^self _basicAt: anIndex
%

category: 'Updating'
method: RcIndexDictionary
basicAt: anIndex put: aValue

"Protected version of Object | basicAt:put:"
<protected>

^self _basicAt: anIndex put: aValue
%

category: 'Clustering'
method: RcIndexDictionary
moveToDisk

"Move the collision buckets, then the receiver to disk."
<protected>

1 to: self _basicSize do: [ :i |
    (self _at: i) moveToDisk
].
super moveToDisk
%

category: 'Reduced Conflict Support'
method: RcIndexDictionary
_resolveRcConflictsWith: conflictObjects

"A logical write-write conflict has occurred on the receiver.  The objects that
 had the actual physical write-write conflicts are in the conflictObjects
 Array.  Selectively abort the receiver and then attempt to replay the
 operations maintained in the System redo log.  Returns true if all the
 operations could be replayed; otherwise returns false."

< unprotected > 

" if any conflict objects are BucketValueBag, abort them "
1 to: conflictObjects size do: [ :i |
    (BucketValueBag _hasInstance: (conflictObjects at: i))
        ifTrue: [ (conflictObjects at: i) _selectiveAbort ]
].

^ super _resolveRcConflictsWith: conflictObjects
%

category: 'Clustering'
method: RcIndexDictionary
clusterDepthFirst

"This method clusters the receiver and its collision buckets.
 Has no effect and returns true if the receiver was previously clustered
 in the current transaction, otherwise returns false."
<protected>

self cluster
  ifTrue: [ ^ true ].

1 to: self _basicSize do: [ :i | (self _at: i) cluster ].
^ false
%

category: 'Statistics'
method: RcIndexDictionary
_statisticsInto: dict

"Puts statistical information into the given dictionary."
<protected>

| pages bktCount count maxBkt maxBktSize collBkt numElemArr tableSizeArr 
  phySize |
pages := IdentitySet new.
numElemArr := Array new.
tableSizeArr := Array new.
phySize := self physicalSize.

bktCount := 0.
count :=0.
maxBktSize := 0.
System _addEntireObjectToRcReadSet: self.

1 to: self _basicSize do: [ :i |
  collBkt := self _at: i.
  System _addRootObjectToRcReadSet: collBkt.

  pages add: collBkt page.

  numElemArr add: collBkt numElements.
  tableSizeArr add: collBkt tableSize.

  collBkt size > 0
    ifTrue: [
      bktCount := bktCount + 1.
      count := count + collBkt size.
      (collBkt size > maxBktSize)
        ifTrue: [
          maxBktSize := collBkt size.
          maxBkt := collBkt
        ]
    ].

  phySize := phySize + collBkt physicalSize.
].

(bktCount > 0)
  ifTrue: [
    dict at: #AvgEntriesPerIndexBucket put: (count asFloat / bktCount) 
  ]
  ifFalse: [ dict at: #AvgEntriesPerIndexBucket put: 0].

dict at: #LargestIndexBucketSize
  put: (maxBkt == nil ifTrue: [ 0 ] ifFalse: [ maxBkt numElements ]).
dict at: #IndexDictionaryRootSize put: self _basicSize.
dict at: #IndexDictionaryPages put: pages.
dict at: #IndexDictionarySpaceUtilization put: #[ numElemArr, tableSizeArr ].
dict at: #IndexDictionaryPhysicalSize put: phySize.
%

category: 'Updating'
method: RcIndexDictionary
_refreshBucketCaches

"Verifies that the bucket caches are correct.  If there is an invalid cache
 entry, fixes it and reports any problems.  Returns a string."
<protected>

| str result |
result := String new.
1 to: self _basicSize do: [ :i |
  str := String new.
  (self _at: i) _refreshBucketCacheReportingInto: str.
  str isEmpty
    ifFalse: [
      result add: 'For index bucket at '; add: i asString;
        addAll: str
    ]
].
^ result
%

category: 'Private'
method: RcIndexDictionary
_canonicalizeSymbolAt: offset oldSymbol: old newSymbol: newSymbol

"Override the method, but do nothing here because other updates will fix it."

<protected>
  ^self
%

category: 'Accessing'
method: RcIndexDictionary
okToUpdate

"Returns the value of the instance variable 'okToUpdate'."

<protected>

^okToUpdate
%

category: 'Updating'
method: RcIndexDictionary
okToUpdate: newValue

"Modifies the value of the instance variable 'okToUpdate'."

<protected>

okToUpdate := newValue
%

category: 'Accessing'
method: RcIndexDictionary
entryHolders

"Returns the value of the instance variable 'entryHolders'."

<protected>

^entryHolders
%

category: 'Updating'
method: RcIndexDictionary
entryHolders: newValue

"Modifies the value of the instance variable 'entryHolders'."

<protected>

entryHolders := newValue
%

category: 'Accessing'
method: RcIndexDictionary
getWorkingSetInterval: nsc

""

<protected>

^ System
  rcValueCacheAt: #indexDictionaryWorkingSetInterval
  for: nsc
  otherwise: RcIndexDictionary defaultWorkingSetInterval.
%

category: 'Accessing'
method: RcIndexDictionary
getWorkingSetThreshold: nsc

""

<protected>

^ System
  rcValueCacheAt: #indexDictionaryWorkingSetThreshold
  for: nsc
  otherwise: RcIndexDictionary defaultWorkingSetThreshold.
%

category: 'Constants'
classmethod: RcIndexDictionary
defaultWorkingSetInterval

"Returns the default working set interval."

<unprotected>

^ 100
%

category: 'Constants'
classmethod: RcIndexDictionary
defaultWorkingSetThreshold

"Returns the index size threshold at which optimized index creation,
using the working sets, is used."

<unprotected>

^ 1000000
%

category: 'Private'
method: RcIndexDictionary
_selectiveAbort

"Private.  Do selective abort in protected mode."

< unprotected primitive: 901>

super _selectiveAbort.
System _disableProtectedMode.
%

category: 'Private'
method: RcIndexDictionary
_at: aKey put: aValue term: pathTerm logging: aBoolean 

"Stores the aKey/aValue/pathTerm triplet in the index dictionary. Rebuilds the 
 hash table if the addition caused the number of collisions to exceed the limit 
 allowed. The argument aBoolean determines if an entry is placed in the redo 
 log for this operation." 

< unprotected primitive: 901> 
| collBkt | 

okToUpdate 
  ifFalse: [ | hash | 
    hash := (aKey identityHash \\ self _basicSize) + 1. 
    " workingSetInterval is the number of collision buckets covered " 
    hash < workingSetInterval 
      ifFalse: [ | offset | 
      " get Array of IndexDictionaryEntryHolders " 
      offset := (hash \\ entryHolders size) + 1. 
      (entryHolders at: offset) 
        addKey: aKey 
        value: aValue 
        term: pathTerm 
        from: entryHolders 
        at: offset. 
      System _disableProtectedMode. 
      ^ true 
    ] 
  ]. 

collBkt := self indexBucketFor: aKey. 
(collBkt size >= collisionLimitPerBucket) 
  ifTrue: [ 
    self rebuildTable. 
    collBkt := self indexBucketFor: aKey. 
  ]. 
collBkt at: aKey put: aValue term: pathTerm root: self logging: aBoolean. 
aBoolean 
  ifTrue: [ 
    self logAt: aKey put: aValue term: pathTerm inCollisionBucket: collBkt 
  ]. 
System _disableProtectedMode. 
^ true 
%

category: 'Repository Conversion'
method: RcIndexDictionary
rehashForConversion

"Private. Rehashes the receiver because the hash values of some of its keys
 may have changed.

 Instances of RcIndexDictionary do not need to be rehashed."

<unprotected>

^ self.
%

unprotectmethods

